import os
import sys
from typing import Sequence

import numpy as np
import torch
import torch.distributed as dist
import torch.multiprocessing as mp

MASTER_IP = '127.0.0.1'
MASTER_PORT = '50001'
TORCH_TYPE_TO_NUMPY_TYPE = {
    torch.float16: np.float16,
    torch.bfloat16: np.float32,
}

if sys.version_info >= (3, 9):
    list_annotation, tuple_annotation = list, tuple
else:
    from typing import List, Tuple
    list_annotation, tuple_annotation = List, Tuple


def is_torch_npu_installed() -> bool:
    if sys.version_info >= (3, 8):
        from importlib import metadata
        return list(metadata.distributions(name='torch_npu'))
    else:
        import subprocess
        result = subprocess.run(['pip', 'show', 'torch_npu'])
        return result.returncode == 0


def init(rank: int, world_size: int) -> None:
    import torch_npu
    torch_npu.npu.set_device(rank)
    init_method = f'tcp://{MASTER_IP}:{MASTER_PORT}'
    dist.init_process_group(backend='hccl', init_method=init_method, world_size=world_size, rank=rank)


def get_hcomm_info(rank: int) -> str:
    group = dist.distributed_c10d._get_default_group()
    if torch.__version__ > '2.0.1':
        return group._get_backend(torch.device('npu')).get_hccl_comm_name(rank)
    else:
        return group.get_hccl_comm_name(rank)


def generate_random_input(shape: tuple_annotation[int, int], dtype: type) -> np.ndarray:
    return np.random.randn(*shape).astype(dtype)


def convert_numpy_array_to_torch_tensor(array: np.ndarray, dtype: torch.dtype) -> torch.Tensor:
    tensor = torch.tensor(array).npu()
    if dtype == torch.bfloat16:
        tensor = tensor.to(torch.bfloat16)
    return tensor


def generate_self_and_x2(self_shape: tuple_annotation[int, int], x2_shape: tuple_annotation[int, int], dtype: torch.dtype) -> tuple_annotation[np.ndarray, np.ndarray]:
    self_ = generate_random_input(self_shape, TORCH_TYPE_TO_NUMPY_TYPE[dtype])
    x2 = generate_random_input(x2_shape, TORCH_TYPE_TO_NUMPY_TYPE[dtype])
    return self_, x2


def convert_torch_tensor_to_numpy_array(tensor: torch.Tensor) -> np.ndarray:
    if tensor.dtype == torch.bfloat16:
        tensor = tensor.to(torch.float32)
    return tensor.cpu().numpy()


def convert_torch_tensors_to_numpy_arrays(tensors: tuple_annotation[torch.Tensor]) -> list_annotation[np.ndarray]:
    return [convert_torch_tensor_to_numpy_array(tensor) for tensor in tensors]


def save_multi_rank_numpy_arrays(arrays: Sequence[np.ndarray], save_path_prefix: str, world_size: int, rank: int) -> None:
    total_num = len(arrays) * world_size
    max_index_digits = len(str(total_num))
    for index, array in enumerate(arrays):
        index_str = str(len(arrays) * rank + index).zfill(max_index_digits)
        np.save(save_path_prefix + index_str + '.npy', array)


def save_data(input_arrays: tuple_annotation[np.ndarray], output_tensors: tuple_annotation[torch.Tensor], save_path_prefix: str, world_size: int, rank: int) -> None:
    output_arrays = convert_torch_tensors_to_numpy_arrays(output_tensors)
    save_multi_rank_numpy_arrays(input_arrays, save_path_prefix + '_input', world_size, rank)
    save_multi_rank_numpy_arrays(output_arrays, save_path_prefix + '_output', world_size, rank)


def run_all_gather_base_mm_every_rank(rank: int, self_shape: tuple_annotation[int, int], x2_shape: tuple_annotation[int, int], dtype: torch.dtype, world_size: int, trans_x2: bool, gather_output: bool) -> None:
    file_path = sys.argv[0]
    prefix = os.path.splitext(file_path)[0]
    if not is_torch_npu_installed():
        print(f'{prefix} is skip.')
        return
    init(rank, world_size)
    hcomm_info = get_hcomm_info(rank)
    self_array, x2_array = generate_self_and_x2(self_shape, x2_shape, dtype)
    self_tensor = convert_numpy_array_to_torch_tensor(self_array, dtype)
    x2_tensor = convert_numpy_array_to_torch_tensor(x2_array, dtype)
    import torch_npu
    output, gather_out = torch_npu.npu_all_gather_base_mm(self_tensor, x2_tensor.t() if trans_x2 else x2_tensor, hcomm_info, world_size, gather_output=gather_output)
    save_data((self_array, x2_array), (output, gather_out), prefix, world_size, rank)


def run_all_gather_base_mm(self_shape: tuple_annotation[int, int], x2_shape: tuple_annotation[int, int], dtype: torch.dtype, world_size: int, trans_x2: bool, gather_output: bool) -> None:
    mp.spawn(run_all_gather_base_mm_every_rank, args=(self_shape, x2_shape, dtype, world_size, trans_x2, gather_output), nprocs=world_size)


def run_mm_reduce_scatter_base_every_rank(rank: int, self_shape: tuple_annotation[int, int], x2_shape: tuple_annotation[int, int], dtype: torch.dtype, world_size: int, trans_x2: bool) -> None:
    file_path = sys.argv[0]
    prefix = os.path.splitext(file_path)[0]
    if not is_torch_npu_installed():
        print(f'{prefix} is skip.')
        return
    init(rank, world_size)
    hcomm_info = get_hcomm_info(rank)
    self_array, x2_array = generate_self_and_x2(self_shape, x2_shape, dtype)
    self_tensor = convert_numpy_array_to_torch_tensor(self_array, dtype)
    x2_tensor = convert_numpy_array_to_torch_tensor(x2_array, dtype)
    import torch_npu
    output = torch_npu.npu_mm_reduce_scatter_base(self_tensor, x2_tensor.t() if trans_x2 else x2_tensor, hcomm_info, world_size)
    save_data((self_array, x2_array), (output,), prefix, world_size, rank)


def run_mm_reduce_scatter_base(self_shape: tuple_annotation[int, int], x2_shape: tuple_annotation[int, int], dtype: torch.dtype, world_size: int, trans_x2: bool) -> None:
    mp.spawn(run_mm_reduce_scatter_base_every_rank, args=(self_shape, x2_shape, dtype, world_size, trans_x2), nprocs=world_size)
